
From: Christoph Lameter <clameter@sgi.com>

This patch adds a new function alloc_zeroed_user_highpage that is then used
in the anonymous page fault handler and in the COW code to allocate zeroed
pages.  The function can be defined per arch to setup special processing
for user pages by defining __HAVE_ARCH_ALLOC_ZEROED_USER_PAGE.  For arches
that do not need to do special things for user pages,
alloc_zeroed_user_highpage is defined to simply do

	alloc_page_vma(GFP_HIGHUSER | __GFP_ZERO, vma, vaddr)

This patch needs to update a number of archs. Wish there was a better way
to do this.

Signed-off-by: Christoph Lameter <clameter@sgi.com>
Signed-off-by: Andrew Morton <akpm@osdl.org>
---

 25-akpm/include/asm-alpha/page.h     |    3 +++
 25-akpm/include/asm-cris/page.h      |    3 +++
 25-akpm/include/asm-h8300/page.h     |    3 +++
 25-akpm/include/asm-i386/page.h      |    3 +++
 25-akpm/include/asm-ia64/page.h      |   10 ++++++++++
 25-akpm/include/asm-m32r/page.h      |    3 +++
 25-akpm/include/asm-m68knommu/page.h |    3 +++
 25-akpm/include/asm-s390/page.h      |    3 +++
 25-akpm/include/asm-x86_64/page.h    |    2 ++
 25-akpm/include/linux/highmem.h      |   11 +++++++++++
 25-akpm/mm/memory.c                  |   32 +++++++++++---------------------
 11 files changed, 55 insertions(+), 21 deletions(-)

diff -puN include/asm-alpha/page.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue include/asm-alpha/page.h
--- 25/include/asm-alpha/page.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue	2005-01-22 23:34:49.556489264 -0800
+++ 25-akpm/include/asm-alpha/page.h	2005-01-22 23:34:49.578485920 -0800
@@ -18,6 +18,9 @@
 extern void clear_page(void *page);
 #define clear_user_page(page, vaddr, pg)	clear_page(page)
 
+#define alloc_zeroed_user_highpage(vma, vaddr) alloc_page_vma(GFP_HIGHUSER | __GFP_ZERO, vma, vmaddr)
+#define __HAVE_ARCH_ALLOC_ZEROED_USER_HIGHPAGE
+
 extern void copy_page(void * _to, void * _from);
 #define copy_user_page(to, from, vaddr, pg)	copy_page(to, from)
 
diff -puN include/asm-cris/page.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue include/asm-cris/page.h
--- 25/include/asm-cris/page.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue	2005-01-22 23:34:49.557489112 -0800
+++ 25-akpm/include/asm-cris/page.h	2005-01-22 23:34:49.578485920 -0800
@@ -21,6 +21,9 @@
 #define clear_user_page(page, vaddr, pg)    clear_page(page)
 #define copy_user_page(to, from, vaddr, pg) copy_page(to, from)
 
+#define alloc_zeroed_user_highpage(vma, vaddr) alloc_page_vma(GFP_HIGHUSER | __GFP_ZERO, vma, vaddr)
+#define __HAVE_ARCH_ALLOC_ZEROED_USER_HIGHPAGE
+
 /*
  * These are used to make use of C type-checking..
  */
diff -puN include/asm-h8300/page.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue include/asm-h8300/page.h
--- 25/include/asm-h8300/page.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue	2005-01-22 23:34:49.558488960 -0800
+++ 25-akpm/include/asm-h8300/page.h	2005-01-22 23:34:49.579485768 -0800
@@ -30,6 +30,9 @@
 #define clear_user_page(page, vaddr, pg)	clear_page(page)
 #define copy_user_page(to, from, vaddr, pg)	copy_page(to, from)
 
+#define alloc_zeroed_user_highpage(vma, vaddr) alloc_page_vma(GFP_HIGHUSER | __GFP_ZERO, vma, vaddr)
+#define __HAVE_ARCH_ALLOC_ZEROED_USER_HIGHPAGE
+
 /*
  * These are used to make use of C type-checking..
  */
diff -puN include/asm-i386/page.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue include/asm-i386/page.h
--- 25/include/asm-i386/page.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue	2005-01-22 23:34:49.560488656 -0800
+++ 25-akpm/include/asm-i386/page.h	2005-01-22 23:34:49.577486072 -0800
@@ -36,6 +36,9 @@
 #define clear_user_page(page, vaddr, pg)	clear_page(page)
 #define copy_user_page(to, from, vaddr, pg)	copy_page(to, from)
 
+#define alloc_zeroed_user_highpage(vma, vaddr) alloc_page_vma(GFP_HIGHUSER | __GFP_ZERO, vma, vaddr)
+#define __HAVE_ARCH_ALLOC_ZEROED_USER_HIGHPAGE
+
 /*
  * These are used to make use of C type-checking..
  */
diff -puN include/asm-ia64/page.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue include/asm-ia64/page.h
--- 25/include/asm-ia64/page.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue	2005-01-22 23:34:49.562488352 -0800
+++ 25-akpm/include/asm-ia64/page.h	2005-01-22 23:34:49.576486224 -0800
@@ -75,6 +75,16 @@ do {						\
 	flush_dcache_page(page);		\
 } while (0)
 
+
+#define alloc_zeroed_user_highpage(vma, vaddr) \
+({						\
+	struct page *page = alloc_page_vma(GFP_HIGHUSER | __GFP_ZERO, vma, vaddr); \
+	flush_dcache_page(page);		\
+	page;					\
+})
+
+#define __HAVE_ARCH_ALLOC_ZEROED_USER_HIGHPAGE
+
 #define virt_addr_valid(kaddr)	pfn_valid(__pa(kaddr) >> PAGE_SHIFT)
 
 #ifdef CONFIG_VIRTUAL_MEM_MAP
diff -puN include/asm-m32r/page.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue include/asm-m32r/page.h
--- 25/include/asm-m32r/page.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue	2005-01-22 23:34:49.563488200 -0800
+++ 25-akpm/include/asm-m32r/page.h	2005-01-22 23:34:49.577486072 -0800
@@ -17,6 +17,9 @@ extern void copy_page(void *to, void *fr
 #define clear_user_page(page, vaddr, pg)	clear_page(page)
 #define copy_user_page(to, from, vaddr, pg)	copy_page(to, from)
 
+#define alloc_zeroed_user_highpage(vma, vaddr) alloc_page_vma(GFP_HIGHUSER | __GFP_ZERO, vma, vaddr)
+#define __HAVE_ARCH_ALLOC_ZEROED_USER_HIGHPAGE
+
 /*
  * These are used to make use of C type-checking..
  */
diff -puN include/asm-m68knommu/page.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue include/asm-m68knommu/page.h
--- 25/include/asm-m68knommu/page.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue	2005-01-22 23:34:49.565487896 -0800
+++ 25-akpm/include/asm-m68knommu/page.h	2005-01-22 23:34:49.578485920 -0800
@@ -30,6 +30,9 @@
 #define clear_user_page(page, vaddr, pg)	clear_page(page)
 #define copy_user_page(to, from, vaddr, pg)	copy_page(to, from)
 
+#define alloc_zeroed_user_highpage(vma, vaddr) alloc_page_vma(GFP_HIGHUSER | __GFP_ZERO, vma, vaddr)
+#define __HAVE_ARCH_ALLOC_ZEROED_USER_HIGHPAGE
+
 /*
  * These are used to make use of C type-checking..
  */
diff -puN include/asm-s390/page.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue include/asm-s390/page.h
--- 25/include/asm-s390/page.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue	2005-01-22 23:34:49.566487744 -0800
+++ 25-akpm/include/asm-s390/page.h	2005-01-22 23:34:49.579485768 -0800
@@ -106,6 +106,9 @@ static inline void copy_page(void *to, v
 #define clear_user_page(page, vaddr, pg)	clear_page(page)
 #define copy_user_page(to, from, vaddr, pg)	copy_page(to, from)
 
+#define alloc_zeroed_user_highpage(vma, vaddr) alloc_page_vma(GFP_HIGHUSER | __GFP_ZERO, vma, vaddr)
+#define __HAVE_ARCH_ALLOC_ZEROED_USER_HIGHPAGE
+
 /* Pure 2^n version of get_order */
 extern __inline__ int get_order(unsigned long size)
 {
diff -puN include/asm-x86_64/page.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue include/asm-x86_64/page.h
--- 25/include/asm-x86_64/page.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue	2005-01-22 23:34:49.567487592 -0800
+++ 25-akpm/include/asm-x86_64/page.h	2005-01-22 23:34:49.577486072 -0800
@@ -38,6 +38,8 @@ void copy_page(void *, void *);
 #define clear_user_page(page, vaddr, pg)	clear_page(page)
 #define copy_user_page(to, from, vaddr, pg)	copy_page(to, from)
 
+#define alloc_zeroed_user_highpage(vma, vaddr) alloc_page_vma(GFP_HIGHUSER | __GFP_ZERO, vma, vaddr)
+#define __HAVE_ARCH_ALLOC_ZEROED_USER_HIGHPAGE
 /*
  * These are used to make use of C type-checking..
  */
diff -puN include/linux/highmem.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue include/linux/highmem.h
--- 25/include/linux/highmem.h~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue	2005-01-22 23:34:49.569487288 -0800
+++ 25-akpm/include/linux/highmem.h	2005-01-22 23:35:52.345943816 -0800
@@ -42,6 +42,17 @@ static inline void clear_user_highpage(s
 	smp_wmb();
 }
 
+#ifndef __HAVE_ARCH_ALLOC_ZEROED_USER_HIGHPAGE
+static inline struct page *
+alloc_zeroed_user_highpage(struct vm_area_struct *vma, unsigned long vaddr)
+{
+	struct page *page = alloc_page_vma(GFP_HIGHUSER, vma, vaddr);
+
+	clear_user_highpage(page, vaddr);
+	return page;
+}
+#endif
+
 static inline void clear_highpage(struct page *page)
 {
 	void *kaddr = kmap_atomic(page, KM_USER0);
diff -puN mm/memory.c~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue mm/memory.c
--- 25/mm/memory.c~alloc_zeroed_user_highpage-to-fix-the-clear_user_highpage-issue	2005-01-22 23:34:49.570487136 -0800
+++ 25-akpm/mm/memory.c	2005-01-22 23:34:49.576486224 -0800
@@ -84,20 +84,6 @@ EXPORT_SYMBOL(high_memory);
 EXPORT_SYMBOL(vmalloc_earlyreserve);
 
 /*
- * We special-case the C-O-W ZERO_PAGE, because it's such
- * a common occurrence (no need to read the page to know
- * that it's zero - better for the cache and memory subsystem).
- */
-static inline void copy_cow_page(struct page * from, struct page * to, unsigned long address)
-{
-	if (from == ZERO_PAGE(address)) {
-		clear_user_highpage(to, address);
-		return;
-	}
-	copy_user_highpage(to, from, address);
-}
-
-/*
  * Note: this doesn't free the actual pages themselves. That
  * has been handled earlier when unmapping all the memory regions.
  */
@@ -1330,11 +1316,16 @@ static int do_wp_page(struct mm_struct *
 
 	if (unlikely(anon_vma_prepare(vma)))
 		goto no_new_page;
-	new_page = alloc_page_vma(GFP_HIGHUSER, vma, address);
-	if (!new_page)
-		goto no_new_page;
-	copy_cow_page(old_page,new_page,address);
-
+	if (old_page == ZERO_PAGE(address)) {
+		new_page = alloc_zeroed_user_highpage(vma, address);
+		if (!new_page)
+			goto no_new_page;
+	} else {
+		new_page = alloc_page_vma(GFP_HIGHUSER, vma, address);
+		if (!new_page)
+			goto no_new_page;
+		copy_user_highpage(new_page, old_page, address);
+	}
 	/*
 	 * Re-check the pte - we dropped the lock
 	 */
@@ -1805,10 +1796,9 @@ do_anonymous_page(struct mm_struct *mm, 
 
 		if (unlikely(anon_vma_prepare(vma)))
 			goto no_mem;
-		page = alloc_page_vma(GFP_HIGHUSER, vma, addr);
+		page = alloc_zeroed_user_highpage(vma, addr);
 		if (!page)
 			goto no_mem;
-		clear_user_highpage(page, addr);
 
 		spin_lock(&mm->page_table_lock);
 		page_table = pte_offset_map(pmd, addr);
_
